"""
Test for creating a simple scitoken.
"""

import os
import sys
import unittest
import tempfile
import shutil
import datetime

# Allow unittests to be run from within the project base.
if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists("../src"):
    sys.path.append("../src")

import jwt
import scitokens
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
import cryptography.hazmat.primitives.asymmetric.ec as ec
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from jwt import DecodeError, InvalidAudienceError
from scitokens.utils.errors import UnsupportedKeyException


class TestCreation(unittest.TestCase):
    """
    Test the creation of a simple SciToken
    """

    def setUp(self):
        self._private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        self._public_key = self._private_key.public_key()
        self._public_pem = self._public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        
        # Force the keycache to create a cache in a new directory
        self.tmp_dir = tempfile.mkdtemp()
        self.old_xdg = os.environ.get('XDG_CACHE_HOME', None)
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir
        # Clear the cache
        self.keycache = scitokens.utils.keycache.KeyCache()
        # Make sure it made the directory where I wanted it
        self.assertTrue(self.keycache.cache_location.startswith(self.tmp_dir))
        self.assertTrue(os.path.exists(self.keycache.cache_location))
        
        self.keycache.addkeyinfo("local", "sample_key", self._private_key.public_key())
        self._token = scitokens.SciToken(key = self._private_key, key_id="sample_key")
        self._no_kid_token = scitokens.SciToken(key = self._private_key)

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        if self.old_xdg:
            os.environ['XDG_CACHE_HOME'] = self.old_xdg
        # Clean up, delete everything

    def test_create(self):
        """
        Test the creation of a simple SciToken.
        """

        token = scitokens.SciToken(key = self._private_key)
        token.update_claims({"test": "true"})
        serialized_token = token.serialize(issuer = "local")

        self.assertEqual(len(serialized_token.decode('utf8').split(".")), 3)
        print(serialized_token)

    def test_ec_create(self):
        """
        Test the creation of a simple Elliptical Curve token
        """
        ec_private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )

        token = scitokens.SciToken(key = ec_private_key, algorithm = "ES256")
        self.assertTrue(isinstance(ec_private_key, ec.EllipticCurvePrivateKey))
        token.update_claims({"test": "true"})
        serialized_token = token.serialize(issuer = "local")

        self.assertEqual(len(serialized_token.decode('utf8').split(".")), 3)
        print(serialized_token)


    def test_ec_public_key(self):
        """
        Test when the public key is provided to deserialize for Elliptical Curve
        """

        ec_private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )
        ec_public_key = ec_private_key.public_key()
        ec_public_pem = ec_public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        token = scitokens.SciToken(key = ec_private_key, algorithm = "ES256")
        serialized_token = token.serialize(issuer = "local")

        new_token = scitokens.SciToken.deserialize(serialized_token, public_key = ec_public_pem, insecure = True)
        self.assertIsInstance(new_token, scitokens.SciToken)

        # With invalid key
        with self.assertRaises(ValueError):
            scitokens.SciToken.deserialize(serialized_token, insecure=True, public_key = "asdf".encode())

        # With a proper key, but not the right one
        private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )
        public_key = private_key.public_key()
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        with self.assertRaises(DecodeError):
            scitokens.SciToken.deserialize(serialized_token, insecure=True, public_key = pem)

    def test_public_key(self):
        """
        Test when the public key is provided to deserialize
        """

        token = scitokens.SciToken(key = self._private_key)
        serialized_token = token.serialize(issuer = "local")

        new_token = scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure = True)
        self.assertIsInstance(new_token, scitokens.SciToken)

        # With invalid key
        with self.assertRaises(ValueError):
            scitokens.SciToken.deserialize(serialized_token, insecure=True, public_key = "asdf".encode())

        # With a proper key, but not the right one
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        with self.assertRaises(DecodeError):
            scitokens.SciToken.deserialize(serialized_token, insecure=True, public_key = pem)




    def test_aud(self):
        """
        Test the audience argument to deserialize
        """
        token = scitokens.SciToken(key = self._private_key)
        token.update_claims({'aud': 'local'})

        serialized_token = token.serialize(issuer = 'local')

        new_token = scitokens.SciToken.deserialize(serialized_token,
                                                   public_key = self._public_pem,
                                                   insecure = True,
                                                   audience = 'local')
        self.assertIsInstance(new_token, scitokens.SciToken)


    def test_any_aud(self):
        """
        Test the audience argument with a ANY token to deserialize

        A token with ANY audience should be accepted by all audiences.
        """
        token = scitokens.SciToken(key = self._private_key)
        token.update_claims({'aud': 'ANY'})

        serialized_token = token.serialize(issuer = 'ANY')

        #with self.assertRaises(InvalidAudienceError):
        #    scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure = True)

        new_token = scitokens.SciToken.deserialize(serialized_token,
                                                   public_key = self._public_pem,
                                                   insecure = True,
                                                   audience = 'https://doesnotexist.edu')
        self.assertIsInstance(new_token, scitokens.SciToken)

    def test_serialize(self):
        """
        Test various edge cases of serialization, particularly around failures.
        """
        with self.assertRaises(NotImplementedError):
            print(self._token.serialize(issuer="local", include_key=True))

        token = scitokens.SciToken()
        with self.assertRaises(scitokens.utils.errors.MissingKeyException):
            print(token.serialize(issuer="local"))

        with self.assertRaises(scitokens.scitokens.MissingIssuerException):
            print(self._token.serialize())

        serialized_token = self._token.serialize(issuer="local")
        self.assertTrue(serialized_token)

        token = scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure=True)
        self.assertTrue(isinstance(token, scitokens.SciToken))

        with self.assertRaises(NotImplementedError):
            print(scitokens.SciToken.deserialize(serialized_token, require_key=True, insecure=True))

        with self.assertRaises(scitokens.scitokens.InvalidTokenFormat):
            print(scitokens.SciToken.deserialize("asdf1234"))

    def test_create_to_validate(self):
        """
        End-to-end test of SciToken creation, verification, and validation.
        """
        self._token['scope'] = "write:/home/example"
        serialized_token = self._token.serialize(issuer="local")
        token = scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure=True)
        enf = scitokens.Enforcer(issuer="local")
        self.assertTrue(enf.test(token, "write", "/home/example/test_file"), msg=enf.last_failure)
        self.assertFalse(enf.test(token, "read", "/home/example/test_file"))
        self.assertFalse(enf.test(token, "write", "/home/other/test_file"))

    def test_multiple_scopes(self):
        """
        End-to-end test of SciToken creation, verification, and validation with multiple scopes.
        """
        self._token['scope'] = "write:/home/example read:/home/read"
        serialized_token = self._token.serialize(issuer="local")
        token = scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure=True)
        enf = scitokens.Enforcer(issuer="local")
        self.assertTrue(enf.test(token, "write", "/home/example/test_file"), msg=enf.last_failure)
        self.assertFalse(enf.test(token, "read", "/home/example/test_file"))
        self.assertFalse(enf.test(token, "write", "/home/other/test_file"))
        self.assertTrue(enf.test(token, "read", "/home/read/test_file"))

    def test_ver(self):
        """
        Testing the version attribute
        """
        self._token['ver'] = 1
        self._token['scope'] = "write:/home/example"
        enf = scitokens.Enforcer(issuer="local")
        self.assertTrue(enf.test(self._token, "write", "/home/example/test_file"))

        # Now set it to a number it shouldn't understand
        self._token['ver'] = 9999
        self.assertFalse(enf.test(self._token, "write", "/home/example/test_file"))

    def test_opt(self):
        """
        Testing the version attribute
        """
        self._token['opt'] = "This is optional information, and should always return true"
        self._token['scope'] = "write:/home/example"
        enf = scitokens.Enforcer(issuer="local")
        self.assertTrue(enf.test(self._token, "write", "/home/example/test_file"))

    def test_contains(self):
        """
        Testing the contains attribute
        """
        self._token['opt'] = "This is optional information, and should always return true"
        self._token['scp'] = "write:/home/example"

        self.assertTrue('opt' in self._token)
        self.assertTrue('scp' in self._token)
        self.assertFalse('notin' in self._token)

    def test_no_kid(self):
        """
        Testing a token without a kid
        """
        serialized_token = self._no_kid_token.serialize(issuer = 'local')
        print(serialized_token)

        # Make sure that without a kid, it throws a value error rather than
        # a key error (there was a bug)
        with self.assertRaises(ValueError):
            token = scitokens.SciToken.deserialize(serialized_token, insecure=True)

        token = scitokens.SciToken.deserialize(serialized_token, public_key = self._public_pem, insecure=True)

    def test_unsupported_key(self):
        """
        Test a token with an unsupported key algorithm
        """
        with self.assertRaises(UnsupportedKeyException):
            scitokens.SciToken(key = self._private_key, algorithm="doesnotexist")

    def test_autodetect_keytype(self):
        """
        Test the autodetection of the key type
        """
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        ec_private_key = ec.generate_private_key(
            ec.SECP256R1(), default_backend()
        )

        # Test when we give it the wrong algorithm type
        with self.assertRaises(scitokens.scitokens.UnsupportedKeyException):
            token = scitokens.SciToken(key = private_key, algorithm="ES256")

        # Test when we give it the wrong algorithm type
        with self.assertRaises(scitokens.scitokens.UnsupportedKeyException):
            token = scitokens.SciToken(key = ec_private_key, algorithm="RS256")

        # Test when we give an unsupported algorithm
        unsupported_private_key = ec.generate_private_key(
            ec.SECP521R1(), default_backend()
        )
        with self.assertRaises(scitokens.scitokens.UnsupportedKeyException):
            token = scitokens.SciToken(key = unsupported_private_key)

        token = scitokens.SciToken(key = ec_private_key, algorithm="ES256")
        token.serialize(issuer="local")


    def test_discover(self):
        """
        Test wlcg bearer token discovery
        """
        # unset any wlcg discovery environment variables
        try:
           del os.environ['BEARER_TOKEN']
        except KeyError:
           pass
        try:
           del os.environ['BEARER_TOKEN_FILE']
        except KeyError:
           pass
        try:
           del os.environ['XDG_RUNTIME_DIR']
        except KeyError:
           pass

        # move any /tmp/bt_u$ID file out of the way
        try:
            bt_file = 'bt_u{}'.format(os.geteuid())
        except AttributeError as exc:  # windows doesn't have geteuid
            self.skipTest(str(exc))
        bt_path = os.path.join('/tmp', bt_file)
        (bt_fd, bt_tmp) = tempfile.mkstemp()
        os.close(bt_fd)
        if os.path.isfile(bt_path):
            shutil.move(bt_path, bt_tmp)

        # check that the function fails properly
        with self.assertRaises(IOError):
            print(self._token.discover())

        # generate a token and save it as /tmp/bt_u$ID
        tmp_file_token = scitokens.SciToken(key = self._private_key, key_id="tmp_file")
        tmp_file_token['scope'] = 'tmp_file'
        tmp_file_token_s = tmp_file_token.serialize(issuer="local")
        with open(bt_path, 'w') as f:
            f.write(tmp_file_token_s.decode('utf-8'))

        # discover a token and check we found /tmp/bt_u$ID
        token = self._token.discover(public_key = self._public_pem)
        self.assertEqual(token._serialized_token, tmp_file_token._serialized_token)

        # generate a token and save it as $XDG_RUNTIME_DIR/bt_u$ID
        xdg_file_token = scitokens.SciToken(key = self._private_key, key_id="xdg_file")
        xdg_file_token['scope'] = 'xdg_file'
        xdg_file_token_s = xdg_file_token.serialize(issuer="local")
        xdg_dir = tempfile.mkdtemp()
        xdg_path = os.path.join(xdg_dir, bt_file)
        with open(xdg_path, 'w') as f:
            f.write(xdg_file_token_s.decode('utf-8'))

        # set the wlcg discovery environment variable
        os.environ['XDG_RUNTIME_DIR'] = xdg_dir

        # discover a token and check we found $XDG_RUNTIME_DIR/bt_u$ID
        # and not /tmp/bt_u$ID
        token = self._token.discover(public_key = self._public_pem, insecure=True)
        self.assertNotEqual(token._serialized_token, tmp_file_token._serialized_token)
        self.assertEqual(token._serialized_token, xdg_file_token._serialized_token)

        # generate a token and save it in BEARER_TOKEN_FILE
        bearer_file_token = scitokens.SciToken(key = self._private_key, key_id="bearer_file")
        bearer_file_token['scope'] = 'bearer_file'
        bearer_file_token_s = bearer_file_token.serialize(issuer="local")
        (fd, bearer_token_file) = tempfile.mkstemp()
        with open(bearer_token_file, 'w') as f:
            f.write(bearer_file_token_s.decode('utf-8'))
        os.close(fd)

        # set the wlcg discovery environment variable
        os.environ['BEARER_TOKEN_FILE'] = bearer_token_file

        # discover a token and check we found BEARER_TOKEN_FILE
        # and not $XDG_RUNTIME_DIR/bt_u$ID or /tmp/bt_u$ID
        token = self._token.discover(public_key = self._public_pem, insecure=True)
        self.assertNotEqual(token._serialized_token, tmp_file_token._serialized_token)
        self.assertNotEqual(token._serialized_token, xdg_file_token._serialized_token)
        self.assertEqual(token._serialized_token, bearer_file_token._serialized_token)

        # generate a token
        bearer_token = scitokens.SciToken(key = self._private_key, key_id="bearer")
        bearer_token['scope'] = 'bearer'
        bearer_token_s = bearer_token.serialize(issuer="local")

        # set the wlcg discovery environment variable
        os.environ['BEARER_TOKEN'] = bearer_token_s.decode('utf-8')

        # discover a token and check we found BEARER_TOKEN
        # and not BEARER_TOKEN_FILE, $XDG_RUNTIME_DIR/bt_u$ID or /tmp/bt_u$ID
        token = self._token.discover(public_key = self._public_pem, insecure=True)
        self.assertNotEqual(token._serialized_token, tmp_file_token._serialized_token)
        self.assertNotEqual(token._serialized_token, xdg_file_token._serialized_token)
        self.assertNotEqual(token._serialized_token, bearer_file_token._serialized_token)
        self.assertEqual(token._serialized_token, bearer_token._serialized_token)

        # clean up the files and directories created
        shutil.rmtree(xdg_dir)
        os.remove(bearer_token_file)
        os.remove(bt_path)
        if os.path.isfile(bt_tmp):
            shutil.move(bt_tmp, bt_path)

    def test_noiss(self):
        """
        Verify that tokens without the `iss` are not accepted.
        """
        encoded_jwt = jwt.encode(
            {
                "ver": "scitoken:2.0",
                "aud": "https://demo.scitokens.org",
                "exp": datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds=30),
                "iat": datetime.datetime.now(tz=datetime.timezone.utc),
                "nbf": datetime.datetime.now(tz=datetime.timezone.utc),
                "jti": "eab04181-b63a-42aa-b77d-804742829fc5",
                "kid": "key-rs256"
            },
            "secret",
            algorithm="HS256",
            headers={"typ": "JWT", "alg": "HS256", "kid": "key-hs256"}
        )

        # Test when we give it without issuer
        with self.assertRaises(scitokens.scitokens.MissingIssuerException):
           scitokens.SciToken.deserialize(encoded_jwt)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python

"""
Create a sample scitoken, signed with elliptic curve cryptography, from a well-known private key.
"""

import jwt
import json
import base64

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.utils import int_to_bytes

def long_to_base64(data, size=None):
    """
    base64-encode a large integer.
    """
    return base64.urlsafe_b64encode(int_to_bytes(data, size)).strip(b'=')

def gen_jwk():
    """
    Return two JSON Web Keys corresponding to a public and private key
    in a keypair.
    """
    private_key = ec.generate_private_key(
        ec.SECP256R1(), default_backend()
    )
    numbers = private_key.private_numbers()
    private_jwk = {"kty": "EC", "crv": "P-256", "d": long_to_base64(numbers.private_value)}
    public_jwk = {"kty": "EC", "crv": "P-256", "x": long_to_base64(numbers.public_numbers.x),
                  "y": long_to_base64(numbers.public_numbers.y)}
    return public_jwk, private_jwk, private_key


def main():
    """
    Main method for testing tool for creating a sample SciToken.
    """

    with open("sample_ecdsa_keypair.pem", "r") as file_pointer:
        serialized_pair = file_pointer.read()

    loaded_public_key = serialization.load_pem_public_key(
        serialized_pair,
        backend=default_backend()
    )
    # Does the public key make sense?
    serialized_public = loaded_public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    print("My public key:\n", serialized_public)

    loaded_private_key = serialization.load_pem_private_key(
        serialized_pair,
        password=None, # Hey, it's a sample file committed to disk...
        backend=default_backend()
    )
    # Echo out the private key:
    serialized_private = loaded_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    print("My private key:\n", serialized_private)

    public_jwk, private_jwk, child_private_key = gen_jwk()

    print("Instance JWK keypair:", public_jwk, private_jwk)

    # Ok, now generate a token and verify it.
    token_encoded = jwt.encode({"read": "/ligo"}, serialized_private, algorithm="ES256",
        headers={"x5u": "https://vo.example.com/JWS", "cwk": public_jwk})
    #child_token_encoded = jwt.encode({"read": "/ligo/brian"}, serialized_child_private, algorithm="ES256",
    #                                 headers={"pwt": pwt})
    signature = token_encoded.split(".")[-1]

    #numbers = loaded_private_key.private_numbers()

    flattened = {}
    flattened['payload'] = jwt.decode(token_encoded)
    flattened['protected'] = jwt.get_unverified_header(token_encoded)
    flattened['signature'] = token_encoded.split(".")[-1]

    print("My encoded token:\n", token_encoded)

    print("Plain-text token:\n", flattened)

    header = jwt.get_unverified_header(token_encoded)
    print("Non-validated header:\n", header)

    token_decoded = jwt.decode(token_encoded, serialized_public, algorithm="ES256")
    print("Validated token:\n", token_decoded)

    # Let's generate the child token
    pwt = {"payload": token_decoded, "protected": header, "signature": signature}

    serialized_child_private = child_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    child_token_encoded = jwt.encode({"read": "/ligo/brian"}, serialized_child_private, algorithm="ES256",
                                     headers={"pwt": pwt})
    flattened = {}
    flattened['payload'] = jwt.decode(child_token_encoded)
    flattened['protected'] = jwt.get_unverified_header(child_token_encoded)
    flattened['signature'] = child_token_encoded.split(".")[-1]
    flattened['key'] = private_jwk

    print("Child token, encoded:\n", child_token_encoded)
    print("Child token with key:\n", child_token_encoded+"."+base64.urlsafe_b64encode(json.dumps(private_jwk)))
    print("Child token headers:\n", jwt.get_unverified_header(child_token_encoded))
    print("Child token, flattened:\n", flattened)

if __name__ == '__main__':
    main()




"""
Verify a SciToken from command-line inputs.
"""

import argparse
import scitokens
from scitokens.utils.errors import InvalidTokenFormat
from scitokens.utils.errors import MissingIssuerException
from scitokens.utils.errors import MissingKeyException
from scitokens.utils.errors import NonHTTPSIssuer
from scitokens.utils.errors import SciTokensException
from scitokens.utils.errors import UnableToCreateCache
from scitokens.utils.errors import UnsupportedKeyException


def add_args():
    """
    Generate the ArgumentParser object for the CLI.
    """
    parser = argparse.ArgumentParser(description='Verify a new SciToken')
    parser.add_argument('token', type=str, nargs=1, help='The serialized string of SciToken')
    parser.add_argument('-v', '--verbose', action='store_true')

    args = parser.parse_args()
    return args


def main():
    """
    Given a serialized SciToken, verify it and return an error message in case of failure.
    """
    args = add_args()
    stoken = None
    try:
        stoken = scitokens.SciToken.deserialize(args.token[0])

        if args.verbose:
            print("Claims:")
            for claim in stoken.claims():
                print("{}".format(claim))

    except MissingKeyException:
        print("No private key is present.")
    except UnsupportedKeyException:
        print("The provided algorithm in the token is not the one supported by SciToken library (RS256, ES256).")
    except MissingIssuerException:
        print("Issuer not specific in claims or as argument.")
    except NonHTTPSIssuer:
        print("Issuer is not over HTTPS. RFC requires it to be over HTTPS.")
    except InvalidTokenFormat:
        print("Serialized token is not a readable format.")
    except UnableToCreateCache as utcce:
        print("Unable to create cache: {}".format(str(utcce)))
    except SciTokensException as scite:
        print("An error raised from SciTokens library while verifying the token: {}".format(str(scite)))
    except Exception as exc:
        print("An error occurred while verifying the token: {}".format(str(exc)))


if __name__ == "__main__":
    main()



"""
Error and Exceptions in the SciTokens library
"""


class SciTokensException(Exception):
    """
    Base class for exceptions in the SciTokens library
    """
    pass

class MissingKeyException(SciTokensException):
    """
    No private key is present.

    The SciToken required the use of a public or private key, but
    it was not provided by the caller.
    """
    pass

class UnsupportedKeyException(SciTokensException):
    """
    Key is present but is of an unsupported format.

    A public or private key was provided to the SciToken, but
    could not be handled by this library.
    """
    pass

class MissingIssuerException(SciTokensException):
    """
    Missing the issuer in the SciToken, unable to verify token
    """
    pass

class NonHTTPSIssuer(SciTokensException):
    """
    Non HTTPs issuer, as required by draft-ietf-oauth-discovery-07
    https://tools.ietf.org/html/draft-ietf-oauth-discovery-07
    """
    pass

class InvalidTokenFormat(SciTokensException):
    """
    Encoded token has an invalid format.
    """
    pass

class UnableToCreateCache(SciTokensException):
    """
    Unable to make the KeyCache
    """



"""
Test the full HTTP to SciToken serialize and deserialize
"""

import os
import sys
import unittest
import tempfile
import shutil
from pathlib import Path

# Allow unittests to be run from within the project base.
if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists("../src"):
    sys.path.append("../src")

import scitokens
import scitokens.scitokens

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from create_webserver import start_server

TESTS_DIR = Path(__file__).parent


class TestDeserialization(unittest.TestCase):
    """
    Test the deserialization of a SciToken
    """

    def setUp(self):
        # Force the keycache to create a cache in a new directory
        self.tmp_dir = tempfile.mkdtemp()
        self.old_xdg = os.environ.get('XDG_CACHE_HOME', None)
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir
             
        with open(TESTS_DIR / 'simple_private_key.pem', 'rb') as key_file:
            self.private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
                )
        self.test_id = "stuffblah"
        self.public_numbers = self.private_key.public_key().public_numbers()

        with open(TESTS_DIR / 'simple_ec_private_key.pem', 'rb') as key_file:
            self.ec_private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
                )
        self.ec_test_id = "ec_test_id"

        self.ec_public_numbers = self.ec_private_key.public_key().public_numbers()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        if self.old_xdg:
            os.environ['XDG_CACHE_HOME'] = self.old_xdg
        # Clean up, delete everything

    def test_deserialization(self):
        """
        Perform the deserialization test
        """

        server_address = start_server(self.public_numbers.n, self.public_numbers.e, self.test_id)
        print(server_address)
        issuer = "http://localhost:{}/".format(server_address[1])
        token = scitokens.SciToken(key=self.private_key, key_id=self.test_id)
        token.update_claims({"test": "true"})
        serialized_token = token.serialize(issuer=issuer)

        self.assertEqual(len(serialized_token.decode('utf8').split(".")), 3)

        scitoken = scitokens.SciToken.deserialize(serialized_token, insecure=True)

        self.assertIsInstance(scitoken, scitokens.SciToken)

        token = scitokens.SciToken(key=self.private_key, key_id="doesnotexist")
        serialized_token = token.serialize(issuer=issuer)
        with self.assertRaises(scitokens.utils.errors.MissingKeyException):
            scitoken = scitokens.SciToken.deserialize(serialized_token, insecure=True)


    def test_ec_deserialization(self):
        """
        Perform the EC deserialization test
        """

        server_address = start_server(self.public_numbers.n, self.public_numbers.e,
                                      self.test_id,
                                      test_ec={'x': self.ec_public_numbers.x,
                                               'y': self.ec_public_numbers.y,
                                               'kid': self.ec_test_id})
        print(server_address)
        issuer = "http://localhost:{}/".format(server_address[1])
        token = scitokens.SciToken(key=self.ec_private_key, key_id=self.ec_test_id,
                                   algorithm="ES256")
        token.update_claims({"test": "true"})
        serialized_token = token.serialize(issuer=issuer)

        self.assertEqual(len(serialized_token.decode('utf8').split(".")), 3)

        scitoken = scitokens.SciToken.deserialize(serialized_token, insecure=True)

        self.assertIsInstance(scitoken, scitokens.SciToken)

        token = scitokens.SciToken(key=self.private_key, key_id="doesnotexist")
        serialized_token = token.serialize(issuer=issuer)
        with self.assertRaises(scitokens.utils.errors.MissingKeyException):
            scitoken = scitokens.SciToken.deserialize(serialized_token, insecure=True)



if __name__ == '__main__':
    unittest.main()



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--no-network", action="store_true", default=False, help="Skip network tests"
    )
    parser.addoption(
        "--no-intensive", action="store_true", default=False, help="Skip intensive tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "network: mark test as needing network")
    config.addinivalue_line("markers", "intensive: mark test as intensive")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--no-network"):
        skip_network = pytest.mark.skip(reason="Need network access to run and --no-network specified")
        for item in items:
            if "network" in item.keywords:
                item.add_marker(skip_network)
        
    if config.getoption("--no-intensive"):
        skip_intensive = pytest.mark.skip(reason="Intensive tests skipped with --no-intensive")
        for item in items:
            if "intensive" in item.keywords:
                item.add_marker(skip_intensive)



"""
Generate a SciToken from command-line inputs (and a local signing key).
"""

import argparse

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend

import scitokens

import scitokens.utils.demo


def add_args():
    """
    Generate the ArgumentParser object for the CLI.
    """
    parser = argparse.ArgumentParser(description='Create a new SciToken')
    parser.add_argument('--demo', action='store_true', help='The payload that needs encrypting')
    parser.add_argument('claims', metavar='C', type=str, nargs='+',
                        help='Claims in the format key=value')
    parser.add_argument('--keyfile',
                        help='Location of the private key file')
    parser.add_argument('--key_id', help='The string key identifier')
    parser.add_argument('--issuer', help="Issuer for the token")
    parser.add_argument('--lifetime', help="Lifetime of the token, in seconds from now. Default: 1200 seconds (20 minutes)",
                        type=int, default=1200)

    args = parser.parse_args()
    return args


def main():
    """
    Given a set of command line parameters, generate a corresponding SciToken.
    """
    args = add_args()
    # If the demo option is called
    if args.demo:
        payload = {}
        for claim in args.claims:
            (key, value) = claim.split('=', 1)
            payload[key] = value
        print(scitokens.utils.demo.token(payload))

    else:
        with open(args.keyfile, "r") as file_pointer:
            private_key_contents = file_pointer.read()

        loaded_private_key = serialization.load_pem_private_key(
            private_key_contents.encode(),
            password=None, # Hey, it's a sample file committed to disk...
            backend=default_backend()
        )

        token = scitokens.SciToken(key=loaded_private_key, key_id=args.key_id)

        for claim in args.claims:
            (key, value) = claim.split('=', 1)
            token.update_claims({key: value})

        serialized_token = token.serialize(issuer=args.issuer, lifetime=args.lifetime)
        print(serialized_token.decode())


if __name__ == "__main__":
    main()



"""
Test for using a configuration file
"""

import os
import unittest
import tempfile
import logging
import scitokens
import scitokens.utils.config
import configparser

class TestConfig(unittest.TestCase):
    """
    Test the configuration parsing
    """

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        scitokens.utils.config.configuration = configparser.ConfigParser(scitokens.utils.config.CONFIG_DEFAULTS)

    def tearDown(self):
        # Clear the config back to defaults each time
        scitokens.set_config()

    def test_config_file(self):
        """
        Test the configuration with a regular config file
        """
        # Get the current directory and pass it the path of test_config.ini
        scitokens.set_config(os.path.join(self.dir_path, "test_config.ini"))

        self.assertEqual(scitokens.utils.config.get("log_file"), "")
        self.assertEqual(scitokens.utils.config.get("log_level"), "DEBUG")

    def test_passing_config(self):
        """
        Test the passing of a configuration parser object
        """
        new_config = configparser.ConfigParser()
        new_config.add_section("scitokens")
        new_config.set("scitokens", "log_level", "WARNING")

        scitokens.set_config(new_config)

        self.assertEqual(scitokens.utils.config.get("log_level"), "WARNING")

    def test_passing_config_log(self):
        """
        Test the with log_file
        """

        new_config = configparser.ConfigParser()
        new_config.add_section("scitokens")
        new_config.set("scitokens", "log_level", "WARNING")

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_file = os.path.join(tmp_dir, "tmp.log")
            new_config.set("scitokens", "log_file", tmp_file)

            scitokens.set_config(new_config)

            self.assertEqual(scitokens.utils.config.get("log_level"), "WARNING")
            self.assertEqual(scitokens.utils.config.get("log_file"), tmp_file)

            # Log a line
            logger = logging.getLogger("scitokens")
            logger.error("This is an error")
            self.assertTrue(os.path.getsize(tmp_file) > 0)

            # close the log files so that TemporaryDirectory can delete itself
            for handler in logger.handlers:
                handler.close()

    def test_no_config(self):
        """
        Test when there is no config
        """

        # This should throw an exception if there is an error
        self.assertEqual(scitokens.utils.config.get("cache_location"), "")



"""
Test demo module
"""

import scitokens.utils.demo
import unittest
import jwt          # to handle jwt exceptions
import time         # to add time delay
import pytest       # to skip tests


class TestToken(unittest.TestCase):

    @pytest.mark.network
    def test_valid_payload(self):
        """
        Test that the token matches the specified payload
        """
        payload = {
            "key1": "val1",
            "key2": "val2"
            }

        token_serialized = scitokens.utils.demo.token(payload)
        try:
            token = scitokens.SciToken.deserialize(token_serialized)            # automatically call verify
        except jwt.exceptions.ImmatureSignatureError:                           # if the token was issued in the future
            print("Token not yet valid. Retrying in 1 second.")
            time.sleep(10)                                                       # add some delay
            token = scitokens.SciToken.deserialize(token_serialized)            # retry
        # assert that the payload is part of the claims
        for key, value in payload.items():
            self.assertIn((key, value), token.claims())

    @pytest.mark.network
    def test_empty_payload(self):
        """
        Test token with empty payload
        """
        payload = {}
        token_serialized = scitokens.utils.demo.token(payload)
        try:
            scitokens.SciToken.deserialize(token_serialized)            # automatically call verify
        except jwt.exceptions.ImmatureSignatureError:
            print("Token not yet valid. Retrying in 1 second.")
            time.sleep(10)
            scitokens.SciToken.deserialize(token_serialized)


class TestParsedToken(unittest.TestCase):
    @pytest.mark.network
    def test_valid_parsed(self):
        """
        Test that the parsed token matches the payload
        """
        payload = {
            "key1": "val1",
            "key2": "val2"
            }

        try:
            token = scitokens.utils.demo.parsed_token(payload)
        except jwt.exceptions.ImmatureSignatureError:
            print("Token was immature, but that might be ok.  Not throwing error, already tested in token")
            return
        for key, value in payload.items():
            self.assertIn((key, value), token.claims())

    @pytest.mark.network
    def test_empty_parsed(self):
        """
        Test token with empty payload
        """
        payload = {}
        try:
            scitokens.utils.demo.parsed_token(payload)
        except jwt.exceptions.ImmatureSignatureError:
            print("Token was immature, but that might be ok.  Not throwing error, already tested in token")



if __name__ == '__main__':
    unittest.main()




"""
Utility to create a simple webserver that will answer oauth .well-known
required requests.  This is so we can test the HTTP requesting part of issuers.
"""

import threading
import json
try:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
except ImportError:
    from http.server import BaseHTTPRequestHandler, HTTPServer

import scitokens.utils

# For use in the HTTP Serve test class
#test_kid = ""
TEST_N = 0
TEST_E = 0
TEST_ID = ""
HTTPD = None
THREAD = None
EC_TEST_ID = ""
EC_TEST_X = 0
EC_TEST_Y = 0

class OauthRequestHandler(BaseHTTPRequestHandler):
    """
    Request handler for the HTTP requests to authenticate deserialization of a SciToken
    """
    def _set_headers(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/json')
        self.send_header('Cache-Control', 'max-age=3600')
        self.end_headers()

    def do_GET(self): # pylint: disable=invalid-name
        """
        Receive the GET command for the oauth certs
        """
        # These values are module-level constants set by start_server();
        # they are only read here so `global` is not needed.  Using
        # the module-level variables directly avoids F824 (unused global)
        # warnings from the linters.
        
        # Make sure the User-Agent is SciTokens*
        user_agent = self.headers.get('User-Agent')
        if not user_agent.startswith("SciTokens"):
            self.send_response(404)
            return
        self._set_headers()
        to_write = ""
        if self.path == "/.well-known/openid-configuration":
            to_write = json.dumps({"jwks_uri": "http://localhost:{}/oauth2/certs".format(HTTPD.server_address[1])})
        elif self.path == "/oauth2/certs":

            # Dummy Key
            dummy_key = {
                'kid': 'dummykey',
                'n': 'reallylongn',
                'e': 'AQAB',
                'alg': "RS256",
                'kty': "RSA"
            }

            key_info = {}
            key_info['kid'] = TEST_ID
            key_info['n'] = scitokens.utils.bytes_from_long(TEST_N).decode('ascii')
            key_info['e'] = scitokens.utils.bytes_from_long(TEST_E).decode('ascii')
            key_info['kty'] = "RSA"
            key_info['alg'] = "RS256"

            if EC_TEST_ID and EC_TEST_X and EC_TEST_Y:
                ec_key_info = {
                    'kid': EC_TEST_ID,
                    'kty': "EC",
                    'crv': "P-256",
                    'x': scitokens.utils.bytes_from_long(EC_TEST_X).decode('ascii'),
                    'y': scitokens.utils.bytes_from_long(EC_TEST_Y).decode('ascii')
                }
                to_write = json.dumps({'keys': [dummy_key, key_info, ec_key_info]})
            else:
                to_write = json.dumps({'keys': [dummy_key, key_info]})

        self.wfile.write(to_write.encode())


def start_server(test_n, test_e, test_id, test_ec = None):
    """
    Man this is ugly.  But you have to set global variables because it's
    impossible to send arguments to the HTTPServer, since you pass the HTTPServer
    as a type, not an instance.

    :param long int test_n: N for an RSA key
    :param long int test_e: E for an RSA key
    :param str test_id: Key ID for the test key
    :param dict test_ec: If you would like to test EC, then set the data structure to:
        kid, x, y
    """
    global TEST_N
    global TEST_E
    global TEST_ID
    global THREAD
    global HTTPD
    global EC_TEST_ID
    global EC_TEST_X
    global EC_TEST_Y

    TEST_N = test_n
    TEST_E = test_e
    TEST_ID = test_id

    if test_ec:
        EC_TEST_ID = test_ec['kid']
        EC_TEST_X = test_ec['x']
        EC_TEST_Y = test_ec['y']

    server_address = ('127.0.0.1', 0)
    HTTPD = HTTPServer(server_address, OauthRequestHandler)
    THREAD = threading.Thread(target=HTTPD.serve_forever)
    THREAD.daemon = True
    THREAD.start()
    return HTTPD.server_address

def shutdown_server():
    """
    Shutdown the web server
    """
    global THREAD
    global HTTPD
    del THREAD
    del HTTPD




import argparse
from scitokens.utils.keycache import KeyCache

def add_args():
    """
    Generate the ArgumentParser object for the CLI.
    """
    parser = argparse.ArgumentParser(description='Update all tokens in the cache')
    parser.add_argument('-f', '--force', action='store_true', help='Force refresh all tokens')
    args = parser.parse_args()
    return args

def main():
    args = add_args()
    keycache = KeyCache()
    res = keycache.update_all_keys(force_refresh=args.force)
    for i in res:
        print(i)

if __name__ == "__main__":
    main()






"""
Test cases for the Validator and Enforcer classes from the scitokens module.
"""

import os
import sys
import time
import unittest

import cryptography.hazmat.backends
import cryptography.hazmat.primitives.asymmetric.rsa

# Allow unittests to be run from within the project base.
if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists("../src"):
    sys.path.append("../src")

import scitokens


class TestValidation(unittest.TestCase):
    """
    Tests related to the Validator object.
    """

    def test_valid(self):
        """
        Basic unit test coverage of the Validator object.
        """

        def always_accept(value):
            """
            A validator that accepts any value.
            """
            if value or not value:
                return True

        validator = scitokens.Validator()
        validator.add_validator("foo", always_accept)

        token = scitokens.SciToken()
        token["foo"] = "bar"

        self.assertTrue(validator.validate(token))
        self.assertTrue(validator(token))


class TestEnforcer(unittest.TestCase):
    """
    Unit tests for the SciToken's Enforcer object.
    """

    _test_issuer = "https://scitokens.org/unittest"

    @staticmethod
    def always_accept(value):
        if value or not value:
            return True

    def setUp(self):
        """
        Setup a sample token for testing the enforcer.
        """
        now = time.time()
        private_key = cryptography.hazmat.primitives.asymmetric.rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=cryptography.hazmat.backends.default_backend()
        )
        self._token = scitokens.SciToken(key=private_key)
        self._token["foo"] = "bar"
        self._token["iat"] = int(now)
        self._token["exp"] = int(now + 600)
        self._token["iss"] = self._test_issuer
        self._token["nbf"] = int(now)

        # Scitoken v2
        self._token2 = scitokens.SciToken(key=private_key)
        self._token2["ver"] = "scitoken:2.0"
        self._token2["foo"] = "bar"
        self._token2["iat"] = int(now)
        self._token2["exp"] = int(now + 600)
        self._token2["iss"] = self._test_issuer
        self._token2["nbf"] = int(now)
        self._token2['wlcg.groups'] = ['groupA', 'groupB']
        self._token2["aud"] = "ANY"


    def test_enforce_v2(self):
        """
        Test the Enforcer object for profile scitokens:2.0
        Also, there is a non-validated attribute, foo.  In 1.0, non-validated attributes
        cause a validation error.  In 2.0, they are ignored.
        """
        with self.assertRaises(scitokens.scitokens.EnforcementError):
            print(scitokens.Enforcer(None))

        # No audience specified
        enf = scitokens.Enforcer(self._test_issuer)
        self.assertFalse(enf.test(self._token2, "read", "/"), msg=enf.last_failure)
        self._token2["scp"] = "read:/"
        self.assertFalse(enf.test(self._token2, "read", "/"), msg=enf.last_failure)

        # Token is set to to ANY, so any audience will work
        enf = scitokens.Enforcer(self._test_issuer, audience = "https://example.unl.edu")
        self._token2["scp"] = "read:/"
        self.assertTrue(enf.test(self._token2, "read", "/"), msg=enf.last_failure)

        # Change the audience from ANY to https://example.com
        self._token2["aud"] = "https://example.com"
        self.assertFalse(enf.test(self._token2, "read", "/"), msg=enf.last_failure)

        # Change back to ANY
        self._token2["aud"] = "ANY"
        self.assertTrue(enf.test(self._token2, "read", "/"), msg=enf.last_failure)

        self._token2["scp"] = "read:/foo/bar"
        self.assertFalse(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)

        self.assertFalse(enf.test(self._token2, "write", "/foo/bar"), msg=enf.last_failure)

        with self.assertRaises(scitokens.scitokens.InvalidPathError):
            print(enf.test(self._token2, "write", "~/foo"))
        
        
    

    def test_v2(self):
        """
        Test the requirements for a v2
        """
        # First, delete the aud
        del self._token2["aud"]
        enf = scitokens.Enforcer(self._test_issuer, audience="https://example.unl.edu")
        self._token2["scope"] = "read:/foo/bar"
        # Should fail, audience is required for 2.0 token
        self.assertFalse(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Now set the audience to ANY
        self._token2["aud"] = "ANY"
        self.assertTrue(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Now set the audience to ANY
        self._token2["aud"] = ["ANY"]
        self.assertTrue(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Now set the audience to ANY
        self._token2["aud"] = ["notathing.com", "ANY"]
        self.assertTrue(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Now to the correct audience
        self._token2["aud"] = "https://example.unl.edu"
        self.assertTrue(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Now to the wrong audience
        self._token2["aud"] = "https://example.com"
        self.assertFalse(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Arbitrary claims are allowed now in v2
        self._token2["madeupclaim"] = "claimsdontmatter"
        self._token2["aud"] = "ANY"
        self.assertTrue(enf.test(self._token2, "read", "/foo/bar"), msg=enf.last_failure)

        # Arbitrary claims should fail in 1.0
        self._token["madeupclaim"] = "claimsdontmatter"
        self._token["aud"] = "ANY"
        self.assertFalse(enf.test(self._token, "read", "/foo/bar"), msg=enf.last_failure)


    def test_enforce(self):
        """
        Test the Enforcer object.
        """
        with self.assertRaises(scitokens.scitokens.EnforcementError):
            print(scitokens.Enforcer(None))

        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self.assertFalse(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        self._token["scp"] = "read:/"
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        enf = scitokens.Enforcer(self._test_issuer, audience = "https://example.unl.edu")
        enf.add_validator("foo", self.always_accept)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        self._token["scp"] = "read:/foo/bar"
        self.assertFalse(enf.test(self._token, "read", "/foo"), msg=enf.last_failure)

        self.assertTrue(enf.test(self._token, "read", "/foo/bar"), msg=enf.last_failure)
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)
        self.assertTrue(enf.test(self._token, "read", "/foo/bar"), msg=enf.last_failure)

        self.assertFalse(enf.test(self._token, "write", "/foo/bar"), msg=enf.last_failure)

        with self.assertRaises(scitokens.scitokens.InvalidPathError):
            print(enf.test(self._token, "write", "~/foo"))

    def test_enforce_scp_path_boundaries(self):
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self._token["scp"] = ["read:/john"]
        self.assertTrue(enf.test(self._token, "read", "/john"), msg=enf.last_failure)
        self.assertTrue(enf.test(self._token, "read", "/john/file"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "/johnathan"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "/johnny"), msg=enf.last_failure)

        self._token["scp"] = ["read:/john/file"]
        self.assertFalse(enf.test(self._token, "read", "/john"), msg=enf.last_failure)

        self._token["scp"] = ["read:/"]
        self.assertTrue(enf.test(self._token, "read", "/arbitrary/path"), msg=enf.last_failure)

        self._token["scp"] = ["read://john"]
        self.assertTrue(enf.test(self._token, "read", "//john//file"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "//johnathan"), msg=enf.last_failure)

    def test_issuer(self):
        """
        Test the issuer claim, with support for multiple valid issuers.
        """
        # issuer argument is required
        with self.assertRaises(scitokens.scitokens.EnforcementError):
            print(scitokens.Enforcer(None))

        self._token["scope"] = "read:/"

        # single issuer match
        enf = scitokens.Enforcer(self._test_issuer, audience="anything")
        enf.add_validator("foo", self.always_accept)
        self.assertEqual(enf._issuer, self._test_issuer)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        # multiple issuer match
        issuers = ["issuer1", self._test_issuer]
        enf = scitokens.Enforcer(issuers)
        enf.add_validator("foo", self.always_accept)
        self.assertListEqual(enf._issuer, issuers)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        # multiple issuer no match
        issuers = ["issuer1", "issuer2"]
        enf = scitokens.Enforcer(issuers)
        self.assertListEqual(enf._issuer, issuers)
        self.assertFalse(enf.test(self._token2, "read", "/"), msg=enf.last_failure)
        self.assertIn("for claim 'iss'", enf.last_failure)

    def test_enforce_scope(self):
        """
        Test the Enforcer object.
        """
        with self.assertRaises(scitokens.scitokens.EnforcementError):
            print(scitokens.Enforcer(None))

        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self.assertFalse(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        self._token["scope"] = "read:/"
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        enf = scitokens.Enforcer(self._test_issuer, audience = "https://example.unl.edu")
        enf.add_validator("foo", self.always_accept)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        self._token["scope"] = "read:/foo/bar"
        self.assertFalse(enf.test(self._token, "read", "/foo"), msg=enf.last_failure)

        self.assertTrue(enf.test(self._token, "read", "/foo/bar"), msg=enf.last_failure)
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)
        self.assertTrue(enf.test(self._token, "read", "/foo/bar"), msg=enf.last_failure)

        self.assertFalse(enf.test(self._token, "write", "/foo/bar"), msg=enf.last_failure)

        with self.assertRaises(scitokens.scitokens.InvalidPathError):
            print(enf.test(self._token, "write", "~/foo"))

    def test_enforce_scope_path_boundaries(self):
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self._token["scope"] = "read:/john"
        self.assertTrue(enf.test(self._token, "read", "/john"), msg=enf.last_failure)
        self.assertTrue(enf.test(self._token, "read", "/john/file"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "/johnathan"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "/johnny"), msg=enf.last_failure)

        self._token["scope"] = "read:/john/file"
        self.assertFalse(enf.test(self._token, "read", "/john"), msg=enf.last_failure)

        self._token["scope"] = "read:/"
        self.assertTrue(enf.test(self._token, "read", "/arbitrary/path"), msg=enf.last_failure)

        self._token["scope"] = "read://john"
        self.assertTrue(enf.test(self._token, "read", "//john//file"), msg=enf.last_failure)
        self.assertFalse(enf.test(self._token, "read", "//johnathan"), msg=enf.last_failure)


    def test_aud(self):
        """
        Test the audience claim
        """
        self._token['scp'] = 'read:/'
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", lambda path : True)
        self._token['aud'] = "https://example.unl.edu"
        self.assertFalse(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        enf = scitokens.Enforcer(self._test_issuer, audience = "https://example.unl.edu")
        enf.add_validator("foo", lambda path : True)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)

    def test_multiple_aud(self):
        """
        Test multiple aud
        """
        self._token['scp'] = 'read:/'

        # Test multiple audiences
        enf = scitokens.Enforcer(self._test_issuer, audience = ["https://example.unl.edu", "https://another.com"])
        enf.add_validator("foo", self.always_accept)
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)
        self._token['aud'] = "https://another.com"
        self.assertTrue(enf.test(self._token, "read", "/"), msg=enf.last_failure)
        self._token['aud'] = "https://doesnotwork.com"
        self.assertFalse(enf.test(self._token, "read", "/"), msg=enf.last_failure)

        self._token2['scope'] = 'read:/'
        self._token2["aud"] = ["https://example.com", "https://another.com"]
        enf = scitokens.Enforcer(self._test_issuer, audience = "https://example.com")
        self.assertTrue(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)
        self._token2["aud"] = ["notexist", "https://another.com"]
        self.assertFalse(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)

        self._token2["aud"] = ["https://example.com", "https://another.com"]
        enf = scitokens.Enforcer(self._test_issuer, audience = ["https://example.com", "blah.com"])
        self.assertTrue(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)
        self._token2["aud"] = ["notexist", "https://another.com"]
        self.assertFalse(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)

        self._token2["aud"] = ["https://example.com", "https://another.com"]
        enf = scitokens.Enforcer(self._test_issuer, audience = ["https://example.com", "blah.com"])
        self.assertTrue(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)
        enf = scitokens.Enforcer(self._test_issuer, audience = ["foo.com", "blah.com"])
        self.assertFalse(enf.test(self._token2, "read", "/foo"), msg=enf.last_failure)

    def test_getitem(self):
        """
        Test the getters for the SciTokens object.
        """
        self.assertEqual(self._token['foo'], 'bar')
        with self.assertRaises(KeyError):
            print(self._token['bar'])
        self.assertEqual(self._token.get('baz'), None)
        self.assertEqual(self._token.get('foo', 'baz'), 'bar')
        self.assertEqual(self._token.get('foo', 'baz', verified_only=True), 'baz')
        self._token.serialize()
        self.assertEqual(self._token['foo'], 'bar')
        self.assertEqual(self._token.get('foo', 'baz'), 'bar')
        self.assertEqual(self._token.get('bar', 'baz'), 'baz')
        self.assertEqual(self._token.get('bar', 'baz', verified_only=True), 'baz')
        self._token['bar'] = '1'
        self.assertEqual(self._token.get('bar', 'baz', verified_only=False), '1')
        self.assertEqual(self._token.get('bar', 'baz', verified_only=True), 'baz')

    def test_gen_acls(self):
        """
        Test the generation of ACLs
        """
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self._token['scope'] = 'read:/'
        acls = enf.generate_acls(self._token)
        self.assertTrue(len(acls), 1)
        self.assertEqual(acls[0], ('read', '/'))

        self._token['scope'] = 'read:/ write:/foo'
        acls = enf.generate_acls(self._token)
        self.assertTrue(len(acls), 2)
        self.assertTrue(('read', '/') in acls)
        self.assertTrue(('write', '/foo') in acls)

        self._token['scope'] = 'read:/foo read://bar write:/foo write://bar'
        acls = enf.generate_acls(self._token)
        self.assertTrue(len(acls), 4)
        self.assertTrue(('read', '/foo') in acls)
        self.assertTrue(('write', '/foo') in acls)
        self.assertTrue(('read', '/bar') in acls)
        self.assertTrue(('write', '/bar') in acls)

        self._token['exp'] = time.time() - 600
        with self.assertRaises(scitokens.scitokens.ClaimInvalid):
            print(enf.generate_acls(self._token))
        self.assertTrue(enf.last_failure)
        self._token['exp'] = time.time() + 600

        self._token['scope'] = 'read:foo'
        with self.assertRaises(scitokens.scitokens.InvalidAuthorizationResource):
            print(enf.generate_acls(self._token))

        self._token['scope'] = 'read'
        with self.assertRaises(scitokens.scitokens.InvalidAuthorizationResource):
            print(enf.generate_acls(self._token))

    def test_sub(self):
        """
        Verify that tokens with the `sub` set are accepted.
        """
        self._token['sub'] = 'Some Great User'
        enf = scitokens.Enforcer(self._test_issuer)
        enf.add_validator("foo", self.always_accept)

        self._token['scope'] = 'read:/'
        acls = enf.generate_acls(self._token)
        self.assertTrue(len(acls), 1)

if __name__ == '__main__':
    unittest.main()



"""
Test the keycache
"""

import os, stat
import sys
import tempfile
import shutil
import unittest
import pytest       # to skip tests
from unittest import mock
from scitokens.utils.keycache import KeyCache
from scitokens.utils.errors import UnableToCreateCache
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

# Python 3 vs. Python 2
try:
    from urllib.error import URLError
except ImportError:
    from urllib2 import URLError

import create_webserver


class TestKeyCache(unittest.TestCase):
    """
    Test the creation of a simple SciToken
    """

    def setUp(self):


        # Force the keycache to create a cache in a new directory
        self.tmp_dir = tempfile.mkdtemp()
        self.old_xdg = os.environ.get('XDG_CACHE_HOME', None)
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir
        # Clear the cache
        self.keycache = KeyCache()

        # make sure it made the directory where I wanted it
        self.assertTrue(self.keycache.cache_location.startswith(self.tmp_dir))
        self.assertTrue(os.path.exists(self.keycache.cache_location))


    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        if self.old_xdg:
            os.environ['XDG_CACHE_HOME'] = self.old_xdg
        # Clean up, delete everything

    @mock.patch("os.makedirs", side_effect=OSError)
    @mock.patch.dict("os.environ")
    def test_cannot_make_cache(self, _):
        """
        Test when the keycache shouldn't be able to make the cache
        """
        os.environ['XDG_CACHE_HOME'] = "/does/not/exists"

        # Make keycache doesn't fail when unable to make cache file
        try:
            keycache = KeyCache()
            del keycache
        except Exception as e:
            self.fail("Creating a cache threw an error, when it should be a silent failure: {}".format(e))


    @unittest.skipIf(sys.platform.startswith("win"), "Test doesn't work on Windows")
    @unittest.skipIf(not sys.platform.startswith("win") and os.getuid() == 0, "Test doesn't work when root")
    def test_cannot_make_cache_permission_denied(self):
        """
        Test when the keycache shouldn't be able to make the cache due to access privilege
        """
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir

        # Limiting access privilege to read-only for the $XDG_CACHE_HOME
        os.chmod(
            self.tmp_dir,
            stat.S_IRUSR |  # Read for user
            stat.S_IRGRP |  # Read for group
            stat.S_IROTH    # Read for other
        )

        # Create a pem encoded public key
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        # Make sure creating a cache and writing/reading to it does not fail
        try:
            keycache = KeyCache()
            self.keycache.addkeyinfo("https://doesnotexists.edu/", "blahstuff", public_key, cache_timer=60)
            del keycache
        except Exception as e:
            self.fail("Creating a cache and writing/reading to it failed: {}".format(e))

        # Revert the access privilege alteration for the $XDG_CACHE_HOME
        os.chmod(
            self.tmp_dir,
            stat.S_IRWXU |  # Read, write, and execute for user
            stat.S_IRWXG |  # Read, write, and execute for group
            stat.S_IRWXO    # Read, write, and execute for other
        )

    def test_empty(self):
        """
        Test when the keycache should be empty
        """
        # Stand up an HTTP server
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_numbers = private_key.public_key().public_numbers()
        test_id = "thisisatestid"
        server_address = create_webserver.start_server(public_numbers.n, public_numbers.e, test_id)
        print(server_address)
        # Now try to get the public key from the server
        pubkey_from_keycache = self.keycache.getkeyinfo("http://localhost:{}/".format(server_address[1]),
                                 test_id,
                                 insecure=True)

        # Now compare the 2 public keys
        public_pem = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        pubkey_pem_from_keycache = pubkey_from_keycache.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.assertEqual(public_pem, pubkey_pem_from_keycache)

        create_webserver.shutdown_server()

    def test_populated(self):
        """
        Test when there should be some entries populated in the sqllite DB
        """
        # Create a pem encoded public key
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.keycache.addkeyinfo("https://doesnotexists.edu/", "blahstuff", public_key, cache_timer=60)

        # Now extract the just inserted key
        pubkey = self.keycache.getkeyinfo("https://doesnotexists.edu/", "blahstuff")

        public_pem2 = pubkey.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.assertEqual(public_pem, public_pem2)

        # Make sure it errors with urlerror when it should not exist
        with self.assertRaises(URLError):
            self.keycache.getkeyinfo("https://doesnotexists.edu/", "asdf")

    @unittest.skipIf(sys.platform.startswith("win"), "Test doesn't work on Windows")
    @unittest.skipIf(not sys.platform.startswith("win") and os.getuid() == 0, "Test doesn't work when root")
    def test_immutable_cache(self):
        """
        Test when there should be some entries populated in the sqllite DB, but the keycache is immutable
        """
        # Create a pem encoded public key
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        # Populate the keycache
        self.keycache.addkeyinfo("https://doesnotexists.edu/", "blahstuff", public_key, cache_timer=60)

        # Make the keycache immutable
        # Limiting access privilege to read-only for the $XDG_CACHE_HOME
        os.chmod(
            self.keycache.cache_location,
            stat.S_IRUSR |  # Read for user
            stat.S_IRGRP |  # Read for group
            stat.S_IROTH    # Read for other
        )

        # Now extract the just inserted key
        pubkey = self.keycache.getkeyinfo("https://doesnotexists.edu/", "blahstuff")
        self.assertIsNotNone(pubkey, "The key should be in the cache")

        # Now try to insert a new key, it should not fail, but it also should not be writable
        self.keycache.addkeyinfo("https://anotherdoesnotexist.edu/", "another", public_key, cache_timer=60)
        # Getting the cache now should fail, but with a URL error, not a reading from the keycache error
        # A URL error because the above addkeyinfo didn't actually add the key to the cache
        # so the keycache tried to download the key from the web, which failed
        with self.assertRaises(URLError):
            another_pubkey = self.keycache.getkeyinfo("https://anotherdoesnotexist.edu/", "another")

        # Revert the access privilege alteration for the $XDG_CACHE_HOME
        os.chmod(
            self.keycache.cache_location,
            stat.S_IRWXU |  # Read, write, and execute for user
            stat.S_IRWXG |  # Read, write, and execute for group
            stat.S_IRWXO    # Read, write, and execute for other
        )

    def test_cache_timer(self):
        """
        Test if the cache max-age is retrieved from the HTTPS resource
        """
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_numbers = private_key.public_key().public_numbers()
        test_id = "thisisatestid"
        server_address = create_webserver.start_server(public_numbers.n, public_numbers.e, test_id)
        print(server_address)

        _, cache_timer = self.keycache._get_issuer_publickey("http://localhost:{}/".format(server_address[1]),
                                            key_id=test_id,
                                            insecure=True)

        self.assertEqual(cache_timer, 345600)
        create_webserver.shutdown_server()

    def test_cache_update_time(self):
        """
        Test if the cache next_update works
        """
        # Create a pem encoded public key
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.keycache.addkeyinfo("https://doesnotexists.edu/", "blahstuff", public_key, cache_timer=60, next_update=-1)

        # Even though the cache is still valid, the next update is triggered
        # We should still get the key, even though the next update fails
        # (invalid url)
        pubkey = self.keycache.getkeyinfo("https://doesnotexists.edu/", "blahstuff")

        public_pem2 = pubkey.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.assertEqual(public_pem, public_pem2)

    def test_cache_update_trigger(self):
        """
        Test when the next_update triggers and goes to the webserver
        """
        # Stand up an HTTP server
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_numbers = private_key.public_key().public_numbers()
        test_id = "thisisatestid"
        server_address = create_webserver.start_server(public_numbers.n, public_numbers.e, test_id)
        print(server_address)

        # Create a pem encoded public key, just to insert, want to make sure
        # it downloads from the server
        tmp_private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = tmp_private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        # Now try to get the public key from the server
        self.keycache.addkeyinfo("http://localhost:{}/".format(server_address[1]),
                                 test_id,
                                 public_key,
                                 cache_timer=60,
                                 next_update=-1)

        # Next update should trigger now
        pubkey_from_keycache = self.keycache.getkeyinfo("http://localhost:{}/".format(server_address[1]),
                                 test_id,
                                 insecure=True)

        # Now compare the 2 public keys
        public_pem = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        pubkey_pem_from_keycache = pubkey_from_keycache.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        self.assertEqual(public_pem, pubkey_pem_from_keycache)

        create_webserver.shutdown_server()


import sqlite3

class TestKeyCacheSQLInjection(unittest.TestCase):
    """
    Regression tests to verify that SQL injection via issuer/key_id is not possible.
    """

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.old_xdg = os.environ.get('XDG_CACHE_HOME', None)
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir
        self.keycache = KeyCache()

        # Generate a test key pair
        self.private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        self.public_key = self.private_key.public_key()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        if self.old_xdg:
            os.environ['XDG_CACHE_HOME'] = self.old_xdg

    def _count_rows(self):
        conn = sqlite3.connect(self.keycache.cache_location)
        curs = conn.cursor()
        curs.execute("SELECT COUNT(*) FROM keycache")
        count = curs.fetchone()[0]
        conn.close()
        return count

    def test_injection_in_issuer_does_not_delete_other_rows(self):
        """
        With the old .format() pattern, an issuer like "x' OR '1'='1" in a
        DELETE would wipe every row. Parameterized queries treat it as a
        literal value, so no rows other than the exact match are affected.
        """
        # Insert a legitimate row
        self.keycache.addkeyinfo("https://legit.example.com/", "key1",
                                 self.public_key, cache_timer=3600)
        self.assertEqual(self._count_rows(), 1)

        # Attempt injection via issuer in addkeyinfo (which DELETEs first)
        malicious_issuer = "x' OR '1'='1"
        self.keycache.addkeyinfo(malicious_issuer, "evil_key",
                                 self.public_key, cache_timer=3600)

        # The legitimate row must still exist, plus the new malicious-literal row
        self.assertEqual(self._count_rows(), 2)

    def test_injection_in_key_id_does_not_delete_other_rows(self):
        """
        A malicious key_id should not be able to affect other rows.
        """
        self.keycache.addkeyinfo("https://legit.example.com/", "key1",
                                 self.public_key, cache_timer=3600)
        self.assertEqual(self._count_rows(), 1)

        malicious_key_id = "x' OR '1'='1"
        self.keycache.addkeyinfo("https://other.example.com/", malicious_key_id,
                                 self.public_key, cache_timer=3600)

        self.assertEqual(self._count_rows(), 2)

    def test_delete_cache_entry_with_injection_string(self):
        """
        _delete_cache_entry with a crafted issuer must not delete unrelated rows.
        """
        self.keycache.addkeyinfo("https://legit.example.com/", "key1",
                                 self.public_key, cache_timer=3600)
        self.assertEqual(self._count_rows(), 1)

        # Try to delete with an injection string — should match nothing
        self.keycache._delete_cache_entry("x' OR '1'='1", "key1")
        self.assertEqual(self._count_rows(), 1)

    def test_union_select_injection_is_literal(self):
        """
        A UNION SELECT payload in the issuer should be stored as a literal
        value, not interpreted as SQL.
        """
        malicious_issuer = "x' UNION SELECT * FROM keycache --"
        self.keycache.addkeyinfo(malicious_issuer, "key1",
                                 self.public_key, cache_timer=3600)
        self.assertEqual(self._count_rows(), 1)

        # The stored issuer should be the literal malicious string
        conn = sqlite3.connect(self.keycache.cache_location)
        curs = conn.cursor()
        curs.execute("SELECT issuer FROM keycache")
        row = curs.fetchone()
        conn.close()
        self.assertEqual(row[0], malicious_issuer)

    def test_getkeyinfo_injection_issuer_no_leak(self):
        """
        getkeyinfo with an injection payload in issuer must not return
        rows belonging to a different issuer.
        """
        self.keycache.addkeyinfo("https://legit.example.com/", "key1",
                                 self.public_key, cache_timer=3600)

        # This injection string would match all rows with the old code
        malicious_issuer = "x' OR '1'='1"
        # getkeyinfo will not find a cached row and will try to download,
        # which will fail — that's expected.  The important thing is it
        # does NOT return the legit key.
        try:
            result = self.keycache.getkeyinfo(malicious_issuer, "key1")
        except Exception:
            result = None
        self.assertIsNone(result)

    def test_negative_cache_with_injection_string(self):
        """
        _add_negative_cache_entry with injection strings stores them literally.
        """
        malicious_issuer = "x' OR '1'='1"
        malicious_key_id = "y' DROP TABLE keycache --"
        self.keycache._add_negative_cache_entry(malicious_issuer, malicious_key_id, 300)
        self.assertEqual(self._count_rows(), 1)

        conn = sqlite3.connect(self.keycache.cache_location)
        curs = conn.cursor()
        curs.execute("SELECT issuer, key_id FROM keycache")
        row = curs.fetchone()
        conn.close()
        self.assertEqual(row[0], malicious_issuer)
        self.assertEqual(row[1], malicious_key_id)


if __name__ == '__main__':
    unittest.main()



from scitokens.utils.keycache import KeyCache

def main():
    # TODO: Make this work
    keycache = KeyCache()
    res = keycache.list_keys()
    header = ["issuer", "expiration", "key_id", "keydata", "next_update"]
    print("{:<30} | {:<19} | {:<35} | {:<20} | {}".format(header[0], header[1], header[2], header[3], header[4]))
    print("-" * 135)
    for record in res:
        print("{:<30} | {:<19} | {:<35} | {:<20} | {}".format(record[0], record[1], record[2], record[3][:20], record[4]))

if __name__ == "__main__":
    main()



#!/usr/bin/python
"""
Unit coverage tests for the urltools sub-package
"""

from scitokens.urltools import unquote, normalize_path

import os
import sys
import unittest

# Allow unittests to be run from within the project base.
if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists("../src"):
    sys.path.append("../src")

class TestUrltools(unittest.TestCase):
    """
    Test various helper functions copied over from the urltools library.
    """

    def test_unquote(self):
        """
        Run through the logic of the unquote function.
        """
        self.assertEqual(unquote('foo%23bar'), 'foo#bar')
        self.assertEqual(unquote('foo%23bar', ['#']), 'foo%23bar')
        with self.assertRaises(TypeError):
            unquote(None)
        self.assertEqual(unquote(""), "")
        self.assertEqual(unquote("abc123"), "abc123")

    def test_normalize_path(self):
        """
        Run through the logic of the normalize_path function.
        """
        self.assertEqual(normalize_path("//////"), "/")
        self.assertEqual(normalize_path("//"), "/")
        self.assertEqual(normalize_path("//foo/bar//baz"), "/foo/bar/baz")
        self.assertEqual(normalize_path("//foo/bar//baz/"), "/foo/bar/baz/")
        self.assertEqual(normalize_path("//f%20oo/bar"), "/f oo/bar")


if __name__ == '__main__':
    unittest.main()




import argparse
from scitokens.utils.keycache import KeyCache

def add_args():
    """
    Generate the ArgumentParser object for the CLI.
    """
    parser = argparse.ArgumentParser(description='Remove a local cached token')
    parser.add_argument('issuer', help='issuer')
    parser.add_argument('key_id', help='key_id')
    parser.add_argument('-f', '--force', action='store_true', help='Force add')
    args = parser.parse_args()
    return args

def main():
    args = add_args()
    keycache = KeyCache()
    res = keycache.add_key(args.issuer, args.key_id, force_refresh=args.force)
    if res != None:
        print("Successfully added token with issuer = {} and key_id = {}!".format(args.issuer, args.key_id))
        print(res)
    else:
        print("Cannot add issuer = {} and key_id = {} !".format(args.issuer, args.key_id))

if __name__ == "__main__":
    main()



"""
Test the admin-create-key tool
"""

import os
import sys
import unittest
from pathlib import Path

# Codacy has issues with subprocess, but this is only in the tests!
import subprocess # nosec
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec
import json
import tempfile

# Allow unittests to be run from within the project base.
if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists("../src"):
    sys.path.append("../src")

from scitokens.utils import long_from_bytes

TESTS_DIR = Path(__file__).parent


class TestKeyCreate(unittest.TestCase):
    """
    Test the admin-create-key tool
    """

    tool = "-m scitokens.tools.admin_create_key"
    to_delete = []

    def setUp(self):
        os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)
        self.to_delete = []

    @staticmethod
    def _test_private(key):
        return serialization.load_pem_private_key(
            key,
            password=None,
            backend=default_backend()
        )

    @staticmethod
    def _test_public(key):
        return serialization.load_pem_public_key(
            key,
            backend=default_backend()
        )

    @staticmethod
    def _test_ec_public_jwk(key):
        """
        Attempt to read in the key into a key object
        """
        keys = json.loads(key.decode('utf-8'))
        public_key_numbers = ec.EllipticCurvePublicNumbers(
            long_from_bytes(keys['keys'][0]['x']),
            long_from_bytes(keys['keys'][0]['y']),
            ec.SECP256R1()
        )
        return public_key_numbers.public_key(default_backend())

    @staticmethod
    def _test_public_jwk(key):
        """
        Attempt to read in the key into a key object
        """
        keys = json.loads(key.decode('utf-8'))
        public_key_numbers = rsa.RSAPublicNumbers(
            long_from_bytes(keys['keys'][0]['e']),
            long_from_bytes(keys['keys'][0]['n'])
        )
        return public_key_numbers.public_key(default_backend())


    @staticmethod
    def _test_ec_private_jwk(key):
        """
        Attempt to read in the key into a private key object
        """
        keys = json.loads(key.decode('utf-8'))
        public_key_numbers = ec.EllipticCurvePublicNumbers(
            long_from_bytes(keys['keys'][0]['x']),
            long_from_bytes(keys['keys'][0]['y']),
            ec.SECP256R1()
        )
        private_key_numbers = ec.EllipticCurvePrivateNumbers(
            long_from_bytes(keys['keys'][0]['d']),
            public_key_numbers
        )
        return private_key_numbers.private_key(default_backend())

    @staticmethod
    def _test_private_jwk(key):
        """
        Attempt to read in the key into a private key object
        """
        keys = json.loads(key.decode('utf-8'))
        public_key_numbers = rsa.RSAPublicNumbers(
            long_from_bytes(keys['keys'][0]['e']),
            long_from_bytes(keys['keys'][0]['n'])
        )
        private_key_numbers = rsa.RSAPrivateNumbers(
            long_from_bytes(keys['keys'][0]['p']),
            long_from_bytes(keys['keys'][0]['q']),
            long_from_bytes(keys['keys'][0]['d']),
            long_from_bytes(keys['keys'][0]['dp']),
            long_from_bytes(keys['keys'][0]['dq']),
            long_from_bytes(keys['keys'][0]['qi']),
            public_key_numbers
        )
        return private_key_numbers.private_key(default_backend())


    def _run_command(self, command):
        # Bandit tests in codacy doesn't like shell=True, but this is suppose to
        # test the user actually running the command from the shell, so keep
        # shell=True is necessary.
        command_run = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE) # nosec
        (stdout, stderr) = command_run.communicate()
        if command_run.returncode != 0:
            print(stdout)
            print(stderr)
            self.assertEqual(command_run.returncode, 0)
        return stdout


    def test_create(self):
        """
        Test the key creation
        """
        command = "{} {} --create-keys --pem-private".format(sys.executable, self.tool)
        output = self._run_command(command)
        private_key = self._test_private(output)
        self.assertIsNotNone(private_key)

        # Test public key
        command = "{} {} --create-keys --pem-public".format(sys.executable, self.tool)
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --create-keys --jwks-private".format(sys.executable, self.tool)
        output = self._run_command(command)
        private_key = self._test_private_jwk(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --create-keys --jwks-public".format(sys.executable, self.tool)
        output = self._run_command(command)
        public_key = self._test_public_jwk(output)
        self.assertIsNotNone(public_key)

    def test_ec_create(self):
        """
        Test the key creation
        """
        command = "{} {} --ec --create-keys --pem-private".format(sys.executable, self.tool)
        output = self._run_command(command)
        private_key = self._test_private(output)
        self.assertIsNotNone(private_key)

        # Test public key
        command = "{} {} --ec --create-keys --pem-public".format(sys.executable, self.tool)
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --ec --create-keys --jwks-private".format(sys.executable, self.tool)
        output = self._run_command(command)
        private_key = self._test_ec_private_jwk(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --ec --create-keys --jwks-public".format(sys.executable, self.tool)
        output = self._run_command(command)
        public_key = self._test_ec_public_jwk(output)
        self.assertIsNotNone(public_key)


    def test_parse_private(self):
        """
        Test reading in the private key
        """
        command = "{} {} --private-key={}/simple_private_key.pem --pem-private".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        private_key = self._test_private(output)
        self.assertIsNotNone(private_key)

        # Test public key
        command = "{} {} --private-key={}/simple_private_key.pem --pem-public".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --private-key={}/simple_private_key.pem --jwks-private".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        private_key = self._test_private_jwk(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --private-key={}/simple_private_key.pem --jwks-public".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        public_key = self._test_public_jwk(output)
        self.assertIsNotNone(public_key)

    def test_ec_parse_private(self):
        """
        Test reading in the private key
        """
        command = "{} {} --ec --private-key={}/simple_ec_private_key.pem --pem-private".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        private_key = self._test_private(output)
        self.assertIsNotNone(private_key)

        # Test public key
        command = "{} {} --ec --private-key={}/simple_ec_private_key.pem --pem-public".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --ec --private-key={}/simple_ec_private_key.pem --jwks-private".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        private_key = self._test_ec_private_jwk(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --ec --private-key={}/simple_ec_private_key.pem --jwks-public".format(
            sys.executable,
            self.tool,
            TESTS_DIR,
        )
        output = self._run_command(command)
        public_key = self._test_ec_public_jwk(output)
        self.assertIsNotNone(public_key)

    def test_parse_public(self):
        """
        Test reading in the public key
        """
        # Create a temporary public key from the private key
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()

        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        tmpfile = tempfile.NamedTemporaryFile(delete=False)

        tmpfile.write(pem)
        tmpfile.close()
        self.to_delete.append(tmpfile.name)

        # Test public key
        command = "{} {} --public-key={} --pem-public".format(sys.executable, self.tool, tmpfile.name)
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --public-key={} --jwks-public".format(sys.executable, self.tool, tmpfile.name)
        output = self._run_command(command)
        public_key = self._test_public_jwk(output)
        self.assertIsNotNone(public_key)

    def test_ec_parse_public(self):
        """
        Test reading in the public key
        """
        # Create a temporary public key from the private key
        private_key = ec.generate_private_key(
            ec.SECP256R1(),
            backend=default_backend()
        )
        public_key = private_key.public_key()

        pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        tmpfile = tempfile.NamedTemporaryFile(delete=False)

        tmpfile.write(pem)
        tmpfile.close()
        self.to_delete.append(tmpfile.name)

        # Test public key
        command = "{} {} --ec --public-key={} --pem-public".format(sys.executable, self.tool, tmpfile.name)
        output = self._run_command(command)
        public_key = self._test_public(output)
        self.assertIsNotNone(public_key)

        # Test public key
        command = "{} {} --ec --public-key={} --jwks-public".format(sys.executable, self.tool, tmpfile.name)
        output = self._run_command(command)
        public_key = self._test_ec_public_jwk(output)
        self.assertIsNotNone(public_key)

    def tearDown(self):
        for file_delete in self.to_delete:
            os.remove(file_delete)
        self.to_delete = []



"""
Test the keycache
"""

import os
import logging
import tempfile
import shutil
import threading
import unittest
from scitokens.utils.keycache import KeyCache
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import pytest
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key


class TestKeyCache(unittest.TestCase):
    """
    Test the creation of a simple SciToken
    """

    def setUp(self):


        # Force the keycache to create a cache in a new directory
        self.tmp_dir = tempfile.mkdtemp()
        self.old_xdg = os.environ.get('XDG_CACHE_HOME', None)
        os.environ['XDG_CACHE_HOME'] = self.tmp_dir
        # Clear the cache
        self.keycache = KeyCache()

        # make sure it made the directory where I wanted it
        self.assertTrue(self.keycache.cache_location.startswith(self.tmp_dir))
        self.assertTrue(os.path.exists(self.keycache.cache_location))


    def tearDown(self):
        import time
        for attempt in range(5):
            try:
                shutil.rmtree(self.tmp_dir)
                break
            except Exception:
                if attempt == 4:
                    raise
                time.sleep(1)
        if self.old_xdg:
            os.environ['XDG_CACHE_HOME'] = self.old_xdg
        # Clean up, delete everything

    @pytest.mark.intensive
    def test_clients_calling_valid_keys(self):
        """
        Test when there are many clients calling valid keys at the same time
        """
        # Thread Job
        # Create a pem encoded public key
        private_key = generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )
        public_key = private_key.public_key()
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

        # Populate the keycache

        def client_job():
            keycache = KeyCache()
            keycache.addkeyinfo("https://doesnotexists.edu/", "blahstuff", public_key, cache_timer=60)
        threads = []
        for _ in range(2000):
            thread = threading.Thread(target=client_job)
            threads.append(thread)
            thread.start()
        for thread in threads:  # iterates over the threads
            thread.join()

    @pytest.mark.network
    @pytest.mark.intensive
    def test_clients_calling_invalid_keys(self):
        """
        Test when there are many clients calling valid keys at the same time
        """
        # Thread Job
        def client_job(issuer, key_id):
            keycache = KeyCache()
            res = keycache.add_key(issuer, key_id, False)
            logger = logging.getLogger("scitokens")
            logger.warning(res)
        threads = []
        key = ('minh', 'vy')
        for _ in range(200):
            thread = threading.Thread(target=client_job, args=key)
            threads.append(thread)
            thread.start()
        for thread in threads:  # iterates over the threads
            thread.join()

if __name__ == '__main__':
    unittest.main()


"""
A module for retrieving a signed token corresponding to a specified payload.
"""

import json
import requests
import scitokens

def token(payload: dict):
    """
    Get a signed token for the given payload.

    :param dict payload: a dictionary specifying the claims
    :returns: an encoded token for the payload
    """
    data = json.dumps({'algorithm': "ES256", 'payload': payload})
    resp = requests.post("https://demo.scitokens.org/issue", data=data)
    return resp.text


def parsed_token(payload: dict):
    """
    Get a parsed token for the given payload.

    :param dict payload: a dictionary specifying the claims
    :returns: a SciToken object
    """
    token = scitokens.utils.demo.token(payload)
    return scitokens.SciToken.deserialize(token)



"""
Install file for SciTokens project.
"""

import os.path
import re

import setuptools

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()


def find_version(path, varname="__version__"):
    """Parse the version metadata variable in the given file.
    """
    with open(path, 'r') as fobj:
        version_file = fobj.read()
    version_match = re.search(
        r"^{0} = ['\"]([^'\"]*)['\"]".format(varname),
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    # metadata
    name="scitokens",
    version=find_version(os.path.join("src", "scitokens", "__init__.py")),
    description="SciToken reference implementation library",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author_email="team@scitokens.org",
    author="Brian Bockelman",
    license="Apache-2.0",
    url="https://scitokens.org",
    project_urls={
        "Documentation": "https://scitokens.readthedocs.io/",
        "Issue Tracker": "https://github.com/scitokens/scitokens/issues",
        "Source Code": "https://github.com/scitokens/scitokens",
    },
    # contents
    package_dir={
        "": "src",
    },
    packages=[
        "scitokens",
        "scitokens.tools",
        "scitokens.utils",
    ],
    entry_points={
        "console_scripts": [
            "scitokens-admin-create-key=scitokens.tools.admin_create_key:main",
            "scitokens-admin-create-token=scitokens.tools.admin_create_token:main",
            "scitokens-verify-token=scitokens.tools.verify_token:main",
        ],
    },
    # requirements
    python_requires=">=3.5",
    install_requires=[
        'cryptography',
        'PyJWT>=1.6.1',
        'requests',
    ],
    extras_require={
        "docs": [
            "Sphinx",
            "sphinx-automodapi",
            "sphinx-rtd-theme",
        ],
    },
    # classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
    ],
)



# This protect decorator is designed to be used with a flask application.  It can be used like:
# @scitokens_protect.protect(audience="https://demo.scitokens.org", scope="read:/secret", issuer="https://demo.scitokens.org”)
# The possible arguments are:
# audience str or list: Audience expected in the client token
# scope str: Scope required to access the function
# issuer str: The issuer to require of the client token
#
# The protected function can optionally take an argument “token”, which is the parsed SciToken object.


import scitokens
from functools import wraps
from flask import request
import traceback
import inspect

def protect(**outer_kwargs):
    def real_decorator(some_function):
        @wraps(some_function)
        def wrapper(*args, **kwargs):
            if 'Authorization' not in request.headers:
                headers = {
                    'WWW-Authenticate': 'Bearer'
                }
                return ("No Authentication Header", 401, headers)
            bearer = request.headers.get("Authorization")
            if len(bearer.split()) != 2:
                headers = {
                    'WWW-Authenticate': 'Bearer'
                }
                return ("Authentication header incorrect format", 401, headers)
            serialized_token = bearer.split()[1]
            try:
                token = scitokens.SciToken.deserialize(serialized_token, audience = outer_kwargs['audience'])
            except Exception as e:
                print(str(e))
                traceback.print_exc()
                headers = {
                    'WWW-Authenticate': 'Bearer'
                }
                return ("Unable to deserialize: %{}".format(str(e)), 401, headers)

            issuers = []
            if not isinstance(outer_kwargs['issuer'], list):
                issuers = [outer_kwargs['issuer']]
            else:
                issuers = outer_kwargs['issuer']
            success = False
            for issuer in issuers:
                enforcer = scitokens.Enforcer(issuer, audience=outer_kwargs['audience'])
                authz, path = outer_kwargs['scope'].split(":")
                if enforcer.test(token, authz, path):
                    success = True
                    break
            if not success:
                headers = {
                    'WWW-Authenticate': 'Bearer'
                }
                return ("Validation incorrect: {}".format(enforcer.last_failure), 403, headers)

            # If the function takes "token" as an argument, send the token
            if 'token' in inspect.getfullargspec(some_function).args:
                kwargs['token'] = token

            return some_function(*args, **kwargs)
        return wrapper
    return real_decorator

@protect(aud="asdf")
def stuff(blah, stuff, **kwargs):
    print(blah)
    print(stuff)
    for key, value in kwargs.iteritems():
        print("%s = %s" % (key, value))




"""
Create the oauth2 certs file from a given private or public key.  The resulting format
should be something like:

{
    "keys": [
        {
            "alg": "RS256",
            "e": "AQAB",
            "kid": "ca78d4d8011b0442025c05885efba3f83764b436961d5cf09cec408ac6c675f2",
            "kty": "RSA",
            "n": "07llFSVsW8cXjy0kG2jHYm084QaSEZcrtw02fHbo30gxgdp6h-maIwNkj_xB-N29kUAe0McoJaL_P4P29rZfh_gh06f9fu60g_GPfVXBuKI61k1FfseaHLtwk2l20WOYHnx92v69UeylJEyNVVYNEhPUHEZdWWqzOjMUIC7XBnQ_GRiGu_9y7JFY-sIS28Iv36r7HKxv1k_i_B5LpOcZ0wUVxmk2WsflgSLg94iUAs8EU2ugI0ea8HG8hP6lPvcWdz4xIQlPYaovrLV_PgTOxVouye979BGIiWgkC7v0hYn5TV7xoIu_3ytm-MKFfas3MZ0cnGBWjlBqVNf-wNguSQ==",
            "use": "sig"
        }
    ]
}

"""

import sys
import cryptography.utils
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, ec
import base64
import binascii
import json
import argparse
from scitokens.utils import string_from_long, bytes_from_long




def add_args():
    """
    Generate the ArgumentParser object for the CLI.
    """
    parser = argparse.ArgumentParser(description='Format a given public key (or create one from a private key) in the format needed for OAuth2 issuer')
    
    # Mutual exclude.  Either give the private keyfile, or public, but not both, and at least 1
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--private-keyfile',
                        help='Location of the unencrypted private key file')
    group.add_argument('--public-keyfile',
                        help='Location of a public key file')
    group.add_argument('--create-keys',
                        help='Create the private and public keys',
                        action='store_true')
    
    parser.add_argument('--jwks-public', help='Print the JWK formatted public key', action='store_true')
    parser.add_argument('--jwks-private', help='Print the JWK formatted private key', action='store_true')
    parser.add_argument('--pem-private', help='Print the PEM formatted private key', action='store_true')
    parser.add_argument('--pem-public', help='Print the PEM formatted public key', action='store_true')
    parser.add_argument('--ec', help='Use eliptical curve cryptograph', action='store_true')

    args = parser.parse_args()
    return args


def main():
    """
    Given a set of command line parameters, generate a corresponding oauth2 certs file.
    """
    args = add_args()
    private_key = None
    public_key = None
    
    
    
    if args.private_keyfile:
        with open(args.private_keyfile, 'rb') as key_file:
            private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )
            public_key = private_key.public_key()
    elif args.public_keyfile:
        # Check for conflicting arguments
        if args.jwks_private or args.pem_private:
            raise Exception("Only given public key on command line, cannot output private keys")
        
        with open(args.public_keyfile, "rb") as key_file:
            public_key = serialization.load_pem_public_key(
                key_file.read(),
                backend=default_backend()
            )
    elif args.create_keys:
        # Create the private key
        if args.ec:
            private_key = ec.generate_private_key(
                ec.SECP256R1(),
                backend=default_backend()
            )
        else:
            private_key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048,
                backend=default_backend()
            )
        public_key = private_key.public_key()
    

    # Get the public numbers
    public_numbers = public_key.public_numbers()

    # Hash the public "n", and use it for the Key ID (kid)
    digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
    if args.ec:
        digest.update(bytes_from_long(public_numbers.x))
    else:
        digest.update(bytes_from_long(public_numbers.n))
    kid = binascii.hexlify(digest.finalize())

    # Shorten the kid to 4 characters
    kid = kid[:4]

    if args.jwks_public:
        if args.ec:
            jwk_public_key = {'keys': [
                {
                    "alg": "ES256",
                    "crv": "P-256",
                    "x": string_from_long(public_numbers.x),
                    "y": string_from_long(public_numbers.y),
                    "kty": "EC",
                    "use": "sig",
                    "kid": kid.decode('utf-8')
                }
            ]}
        else:
            jwk_public_key = {'keys': [
                {
                    "alg": "RS256",
                    "n": string_from_long(public_numbers.n),
                    "e": string_from_long(public_numbers.e),
                    "kty": "RSA",
                    "use": "sig",
                    "kid": kid.decode('utf-8')
                }
            ]}
        print(json.dumps(jwk_public_key, sort_keys=True,
                        indent=4, separators=(',', ': ')))
    
    if args.jwks_private:
        private_numbers = private_key.private_numbers()
        
        if args.ec:
            jwk_private_key = {'keys': [
                {
                    'kty': 'EC',
                    "crv": "P-256",
                    'kid': kid.decode('utf-8'),
                    "x": string_from_long(private_numbers.public_numbers.x),
                    "y": string_from_long(private_numbers.public_numbers.y),
                    'd': string_from_long(private_numbers.private_value)
                }
            ]}
        else:
            jwk_private_key = {'keys': [
                {
                    'kty': 'RSA',
                    "alg":"RS256",
                    'kid': kid.decode('utf-8'),
                    "n": string_from_long(private_numbers.public_numbers.n),
                    "e": string_from_long(private_numbers.public_numbers.e),
                    'p': string_from_long(private_numbers.p),
                    'q': string_from_long(private_numbers.q),
                    'd': string_from_long(private_numbers.d),
                    'dp': string_from_long(private_numbers.dmp1),
                    'dq': string_from_long(private_numbers.dmq1),
                    'qi': string_from_long(private_numbers.iqmp),
                }
            ]}
        print(json.dumps(jwk_private_key, sort_keys=True,
                        indent=4, separators=(',', ': ')))
    
    if args.pem_private:
        private_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )
        for line in private_pem.splitlines():
            print(line.decode())
    
    if args.pem_public:
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        for line in public_pem.splitlines():
            print(line.decode())


if __name__ == "__main__":
    main()



"""
Sample test file for loading a token from command-line parameters.
"""

import scitokens
import sys


token = scitokens.SciToken.deserialize(sys.argv[1].encode())

for claim in token.claims():
    print(claim)

